module compAngles_2CO2_Ar
  use GP_variables_2CO2_Ar, only: lCO
  use AnglesStructures, only: Angles
  use compDist_2CO2_Ar, only: computeCtoO1, computeCtoO3, sqLength
  implicit none

  public computeAngles

  private
  double precision :: smallestNonZeroSinTheta23=1e-7
  type( Angles ), private :: angs
  
contains


  function computeAngles(rab) result( res )
    double precision, dimension(:),intent(in)::  rab
    type( Angles ) :: res

#ifdef DEBUG
    call checkBetasAreLessThanPiBy2( rab )
#endif

    angs%cosTheta23  =  compCosTheta23(rab(5), rab(11), rab(14))
    angs%sinTheta23  =  sqrt( (1.d0 - angs%cosTheta23**2))

    angs%cosBeta1  =   compCosBeta1( rab )
    angs%sinBeta1  =  sqrt(abs(1.d0- angs%cosBeta1**2))

    angs%cosBeta2  =  compCosBeta2( rab )
    angs%sinBeta2  =  sqrt(abs(1.d0- angs%cosBeta2**2))

    angs%cosAlpha1 = compCosAlpha1(rab)
    angs%sinAlpha1 = sqrt( abs(1.d0- angs%cosAlpha1**2))

    if(  abs(angs%sinTheta23) < smallestNonZeroSinTheta23  ) then
       angs%cosAlpha2  =  compCosAlpha2_MoleculesLinear(rab)
    else
       angs%cosAlpha2 = compCosAlpha2(rab)
    end if
    
    angs%sinAlpha2  =  compSinAlpha2(rab)
    angs%alpha2  =  compAlpha2()
   
    angs%beta1 = acos( angs%cosBeta1)
    angs%alpha1 = acos( angs%cosAlpha1)
    angs%beta2 = acos( angs%cosBeta2)

#ifdef DEBUG
    call checkAngles()
#endif

    res = angs
  end function computeAngles
  
  
  
  double precision function compCosTheta23( r12, r13, r23)
    double precision, intent(in):: r12,r13,r23
     
    compCosTheta23 = 0.5d0*r12/r13  + 0.5d0*r13/r12  -  0.5d0*r23**2/r12/r13
    compCosTheta23 = fixCosSinRange( compCosTheta23 )
    
  end function compCosTheta23

  
  
  double precision function compCosBeta1( rab )
    double precision, dimension(:),intent(in)::  rab

    compCosBeta1= (rab(5)**2 + lCO**2 - (min(rab(2),rab(8)))**2)  /  lCO/2.d0/rab(5)
    if( compCosBeta1<0.and.compCosBeta1>-1d-6) then
       compCosBeta1 = 0.d0
    endif

    compCosBeta1 = fixCosSinRange( compCosBeta1 )
  end function compCosBeta1


  
  double precision function compCosBeta2(rab)
    double precision, dimension(:),intent(in)::  rab
    
    compCosBeta2 = max ( (rab(4)**2 - lCO**2 - rab(5)**2 ) , &
         (rab(6)**2 - lCO**2 - rab(5)**2 ) )  /  lCO/2.d0/rab(5)
    
    if(  compCosBeta2 < 0  .and.  compCosBeta2 > -1e-6  ) then
       compCosBeta2 = 0.d0
    endif

    compCosBeta2 = fixCosSinRange( compCosBeta2 )  
  end function compCosBeta2
  
    

  double precision function compCosAlpha1(rab)
    double precision, dimension(:),intent(in)::  rab

    if(angs%sinBeta1 * angs%sinTheta23<1d-12 .or.   &
         abs(angs%sinTheta23)< smallestNonZeroSinTheta23) then
       compCosAlpha1 = 1.d0
    else
       compCosAlpha1 = ((lCO*lCO  -  rab(10)**2 + rab(11)**2) /2.d0/lCO/rab(11) &
            -  angs%cosBeta1 * angs%cosTheta23) /angs%sinBeta1 /angs%sinTheta23
    end if
    
    compCosAlpha1 = fixCosSinRange( compCosAlpha1 )
  end function compCosAlpha1

  

  double precision function compCosTheta14( rab)
    double precision, dimension(:),intent(in)::  rab
    
    compCosTheta14 = rab(11) /rab(14) * angs%sinTheta23
    
    compCosTheta14 = fixCosSinRange( compCosTheta14 )
  end function compCosTheta14


  double precision function compCosAlpha2(rab)
    double precision, dimension(:),intent(in)::  rab
    double precision cosTheta14, sinTheta14

    !==== for alpha2 first compute the angle rab(14) makes with the x axis====
    cosTheta14 = compCosTheta14( rab)
    sinTheta14 = sqrt(abs( 1.d0 - cosTheta14**2))


    if(  abs( angs%sinBeta2 * cosTheta14 ) < 1d-14  ) then
       compCosAlpha2= 1.d0
       return
    endif


    if(  rab(11) * angs%cosTheta23  <=  rab(5)  ) then
       compCosAlpha2 = ((rab(15)**2 - rab(13)**2)/4.d0/lCO/rab(14) &
            +  angs%cosBeta2 * sinTheta14) /angs%sinBeta2 /cosTheta14
    else
       compCosAlpha2 =((lCO*lCO  -  rab(13)**2 + rab(14)**2) /2.d0/lCO/rab(14) &
            -  angs%cosBeta2*sinTheta14) /cosTheta14 /angs%sinBeta2
    endif

    compCosAlpha2 = fixCosSinRange( compCosAlpha2 )
  end function compCosAlpha2


  double precision function compCosAlpha2_MoleculesLinear(rab)
    double precision, dimension(:),intent(in)::  rab
    double precision r12, r1Sq, r2Sq

    if( abs( angs%sinBeta1 * angs%sinBeta2)< 1d-14) then
       compCosAlpha2_MoleculesLinear= 1.d0
       return
    endif

    r12 = rab(5)
    r1Sq = rab(1)**2
    r2Sq = rab(3)**2

    compCosAlpha2_MoleculesLinear= (4.d0 * angs%cosBeta2 * lCO *(-angs%cosBeta1 * lCO + r12)&
         - r1Sq + r2Sq)  &
         /(4.d0 * lCO**2 * angs%sinBeta1 * angs%sinBeta2)

    compCosAlpha2_MoleculesLinear=fixCosSinRange(compCosAlpha2_MoleculesLinear)
  end function compCosAlpha2_MoleculesLinear

  
  double precision function compSinAlpha2(rab)
    double precision, dimension(:),intent(in)::  rab
    double precision positive_root,  r1_sq_pos_root, r1_sq_neg_root, actual_r1_sq

    !Compare the two predicted values for rab(1) (O1->O3) to choose between +ve and -ve roots
    positive_root  = sqrt(  abs( 1.d0 - angs%cosAlpha2**2 )  )
    r1_sq_pos_root = checkR1( rab(5),positive_root)
    r1_sq_neg_root = checkR1( rab(5), -positive_root)

    actual_r1_sq = rab(1)**2
    if( abs(r1_sq_pos_root - actual_r1_sq) <= abs(r1_sq_neg_root - actual_r1_sq)) then
       compSinAlpha2 = positive_root
    else
       compSinAlpha2 = -positive_root 
    endif
  end function compSinAlpha2


  double precision function compAlpha2()
    double precision, parameter :: PI=4.d0*DATAN(1.D0)

    compAlpha2 = acos( angs%cosAlpha2) 
    if( angs%sinAlpha2 < 0.d0) then
       compAlpha2=  2.d0*PI - compAlpha2
    endif
  end function compAlpha2

    
  double precision function checkR1(r12, sinAlpha2Trial)
    double precision, intent(in):: r12,sinAlpha2Trial
    type( Angles ) :: angsTemp
    double precision  O1(3), O3(3)

    O1 = computeCtoO1( angs )

    angsTemp = angs
    angsTemp%sinAlpha2 = sinAlpha2Trial
    O3 = computeCtoO3( angsTemp ) + (/ 0.d0, 0.d0, r12 /)

    checkR1= sqLength(O1,O3)
    
  end function checkR1

  
  subroutine checkBetasAreLessThanPiBy2(rab)
    double precision, dimension(:),intent(in)::  rab
    if( rab(2) > rab(8) ) then ! check O1->C2 smaller than O2->c2
       print *,rab
       print *,"Is beta1 > pi/2  ??"
       stop 100
    endif

    if( rab(6) > rab(4) ) then ! check O3->C1 smaller than O4->C1
       print *,rab
       print *,"Is beta2 > pi/2  ??"
       stop 100
    endif

  end subroutine checkBetasAreLessThanPiBy2

  
  double precision function fixCosSinRange( cosSin )
    double precision, parameter :: cosSinTol= 1d20 !Basically correcting everything!!!
    double precision, intent(in):: cosSin

    fixCosSinRange = cosSin
    
    if (cosSin > 1.d0 .and. cosSin-1.d0 < cosSinTol) then
       fixCosSinRange = 1.d0
    endif
    
    if (cosSin < -1.d0 .and. abs(cosSin+1.d0) < cosSinTol) then
       fixCosSinRange = -1.d0
    endif
    
  end function fixCosSinRange

  
  subroutine checkAngles()
    if(angs%cosBeta1<0.d0 .or. angs%cosBeta2<0.d0 .or. angs%cosBeta1>1.d0 &
         .or. angs%cosBeta2>1.d0) then
       print *,'Problem with one of the cosBetas (should both be positive)'
       print *, angs%cosBeta1, angs%cosBeta2
       !ERROR STOP
    endif
    
    if(angs%cosAlpha1<-1.d0 .or.  angs%cosAlpha1>1.d0 .or. angs%cosAlpha2<-1.d0 &
         .or.angs%cosAlpha2>1.d0 &
         .or. angs%sinAlpha2>1.d0 .or. angs%sinAlpha2 < -1.d0) then
       print *,'Problem with one of the alphas (should both be mod<1)' &
            ,angs%cosAlpha1,angs%cosAlpha2, angs%sinAlpha2
       print *,'Problem with one of the alphas (should both be mod<1)', &
            abs(angs%cosAlpha1)-1.0,abs(angs%cosAlpha2)-1.0, abs(angs%sinAlpha2)-1.0
       !ERROR STOP
    endif
  end subroutine checkAngles


  
end module compAngles_2CO2_Ar
