module permutations
  use Assert_Module
  use GP_variables_2CO2_Ar, only: nDim, distTol
  implicit none
  private



  !!====Public interface================================================
  public fixPermutations
  !!====================================================================

  !!====Public variables/ functions - for tests only====================  
  public flipFirstCO2, flipSecondCO2, swapCO2s
  !!====================================================================

contains


  function fixPermutations( rab ) result (rOut)
    double precision, dimension(nDim), intent(in) :: rab
    double precision, dimension(nDim) :: rOut

    rOut = rab
    rOut = swapAndFlips( rOut )
    rOut = correctEqualCases( rOut )

    call assertLessThanOrEqual_double( rOut(11),rOut(14),distTol,"r23 < r13 after permutation" )
    call assertLessThanOrEqual_double( rOut(2),rOut(8),distTol,"beta1 > pi/2 after permutation")
    call assertLessThanOrEqual_double( rOut(6),rOut(4),distTol,"beta2 > pi/2 after permutation")
    
  end function fixPermutations


  function swapAndFlips( rab ) result (rOut)
    double precision, dimension(nDim), intent(in) :: rab
    double precision, dimension(nDim) :: rOut

    rOut = rab

    if(rOut(11) > rOut(14)) then !r13 vs r23
       rOut = swapCO2s( rOut )
    endif

    if( rOut(2) > rOut(8)) then   !O1->C2 vs O2->C2
       rOut = flipFirstCO2( rOut )
    endif

    if( rOut(6) > rOut(4)) then   !O4->C1 vs O3->C1
       rOut = flipSecondCO2( rOut )
    endif

  end function swapAndFlips


  function correctEqualCases( rab ) result (rOut)
    double precision, dimension(nDim), intent(in) :: rab
    double precision, dimension(nDim) :: rOut
    
    rOut = rab
    if(  R13EqualsR23( rOut)  ) then 
       rOut = correctEqualCtoAr( rOut )
    endif

    if( beta2EqualsPi( rOut )  .and.  O4pointsRight( rOut )  ) then 
       rOut = flipSecondCO2( rOut )
    endif

    if( beta1EqualsPi( rOut )  .and. O2pointsRight( rOut )  ) then 
       rOut = flipFirstCO2( rOut )
    endif
  end function correctEqualCases


  function flipFirstCO2( rab ) result (rOut)
    double precision, dimension(nDim), intent(in) :: rab
    double precision, dimension(nDim) :: rOut

    rOut = rab

    rOut(1) = rab(7)
    rOut(2) = rab(8)
    rOut(3) = rab(9)

    rOut(7) = rab(1)
    rOut(8) = rab(2)
    rOut(9) = rab(3)
    rOut(10) = rab(12)
    rOut(11) = rab(11)
    rOut(12) = rab(10)

  end function flipFirstCO2


  function flipSecondCO2( rab ) result (rOut)
    double precision, dimension(nDim), intent(in) :: rab
    double precision, dimension(nDim) :: rOut

    rOut = rab

    rOut(1) = rab(3)

    rOut(3) = rab(1)
    rOut(4) = rab(6)

    rOut(6) = rab(4)
    rOut(7) = rab(9)

    rOut(9) = rab(7)

    rOut(13) = rab(15)
    rOut(14) = rab(14)
    rOut(15) = rab(13)

  end function flipSecondCO2


  function swapCO2s( rab ) result (rOut)
    double precision, dimension(nDim), intent(in) :: rab 
    double precision, dimension(nDim) :: rOut

    rOut = rab

    rOut(2) = rab(4)
    rOut(3) = rab(7)
    rOut(4) = rab(2)

    rOut(6) = rab(8)
    rOut(7) = rab(3)
    rOut(8) = rab(6)
    rOut(9) = rab(9)
    rOut(10) = rab(13)
    rOut(11) = rab(14)
    rOut(12) = rab(15)
    rOut(13) = rab(10)
    rOut(14) = rab(11)
    rOut(15) = rab(12)

  end function swapCO2s


  function correctEqualCtoAr( rab ) result (rOut)
    double precision, dimension(nDim), intent(in) :: rab
    double precision, dimension(nDim) :: rOut
    
    rOut = rab

    !smallest O-Ar distance should be first CO2
    if( min(rOut(10),rOut(12)) > min(rOut(13) , rOut(15)) ) then
       rOut = swapCO2s( rOut )
    endif

    if( rOut(2) > rOut(8)) then  !O1 should be closer to C2 than O2
       rOut = flipFirstCO2( rOut )
    endif

    if( rOut(6) > rOut(4)) then   !O3 should be closest to C1 than O4
       rOut = flipSecondCO2( rOut )
    endif

  end function correctEqualCtoAr


  
  logical function R13EqualsR23( rOut )
    double precision, dimension(nDim), intent(in) :: rOut
    R13EqualsR23 = .false.
    
    if(abs(rOut(11) - rOut(14)) < distTol ) then 
       R13EqualsR23 = .true.
    endif
  end function R13EqualsR23

  
  logical function beta1EqualsPi( rOut )
    double precision, dimension(nDim), intent(in) :: rOut
    beta1EqualsPi = .false.
    
    if(abs(rOut(2) - rOut(8)) <  distTol ) then 
       beta1EqualsPi = .true.
    endif
  end function beta1EqualsPi

  
  logical function beta2EqualsPi( rOut )
    double precision, dimension(nDim), intent(in) :: rOut
    beta2EqualsPi = .false.
    
    if(abs(rOut(4) - rOut(6)) <  distTol ) then 
       beta2EqualsPi = .true.
    endif
  end function beta2EqualsPi

  
  logical function O4pointsRight( rOut )
    double precision, dimension(nDim), intent(in) :: rOut
    O4pointsRight = .false.

    if( rOut(1) > rOut(3) ) then 
       O4pointsRight = .true.
    endif
  end function O4pointsRight

  
  logical function O2pointsRight( rOut )
    double precision, dimension(nDim), intent(in) :: rOut
    O2pointsRight = .false.

    if( rOut(1) > rOut(7) ) then 
       O2pointsRight = .true.
    endif
  end function O2pointsRight


end module Permutations
