module RegionA_Extrapolate_Module
  use Assert_Module
  use compAngles_2CO2_Ar, only:computeAngles
  use RegionA_Options_2CO2_Ar !Data structure specific to this region so need all variables
  use GP_variables_2CO2_Ar, only: nDim,nPerms
  use PES_GP_Module, only: PES_GP
  use GP_variables_Asympt_2CO2_Ar !Data structure for asymptotes, need all variables
  use compDist_2CO2_Ar, only: computeDistances, compX, compThetaHat
  use CentreToCentreTuple_Module, only: CentreToCentreTuple
  use AnglesStructures, only: Angles
  implicit none

  !!====Public interface================================================
  public RegionA_Extrapolate
  !!====================================================================

  !!====Public variables/ functions - for tests only====================  
  double precision, public::  cosThetaHat_A
  
  public Energy_GP_RegionA, getDistancesAfterSlide_RegionA
  !!====================================================================
  
  private
  type( Angles ), private :: regA_Angles

contains

double precision function RegionA_Extrapolate( rab )
  implicit none
  double precision, intent(in), dimension(nDim) :: rab
  double precision , dimension(nDim) :: xStar
  
  if( minval(rab)<2.d0) then! return the repulsive part
     RegionA_Extrapolate = 0.0   
     return
  endif
  
  if( isInsideGeometricConstrantRegionA(rab) ) then
     xStar(:) = 1/rab(:)
     RegionA_Extrapolate = PES_GP( xStar)
     return
  endif

  RegionA_Extrapolate = RegionA_Extrapolate_WithSlide( rab )
  
end function RegionA_Extrapolate


logical function isInsideGeometricConstrantRegionA( rab )
  implicit none
  double precision, intent(in), dimension(:) :: rab

  if(rab(5)< rShortMax .and. rab(11)< RMedMax .and. rab(14)< RLongMax) then
     isInsideGeometricConstrantRegionA = .true.
  else
     isInsideGeometricConstrantRegionA = .false.
  end if
end function isInsideGeometricConstrantRegionA


double precision function RegionA_Extrapolate_WithSlide( rab)
  implicit none
  double precision, intent(in), dimension(:) :: rab
  double precision, dimension(nDim) :: rabAfterSlide

  rabAfterSlide = getDistancesAfterSlide_RegionA(rab)
  
  RegionA_Extrapolate_WithSlide  =  energyWithScaleAndSlide(rab , rabAfterSlide)

end function RegionA_Extrapolate_WithSlide


function getDistancesAfterSlide_RegionA( rab )  result( rabAfterSlide )
  implicit none
  double precision, intent(in), dimension(:) :: rab
  double precision, dimension(nDim):: rabAfterSlide
  double precision Distace3ToCentreOf12
  type(CentreToCentreTuple) :: rnew

  regA_Angles = computeAngles( rab)

  Distace3ToCentreOf12= compX( rab(5), rab(11), rab(14))
  cosThetaHat_A = compThetaHat(rab(5), rab(11), rab(14), Distace3ToCentreOf12 )
  
  
  rnew = CentreToCentreTuple( rab(5) , rab(11) , rab(14) )

  if( rab(11) > RMedMax) then
     rnew = slideR13toRMedMax_RegionA(rab(5))
  endif

  if( rnew%r23 > RLongMax) then
     rnew = scaleThenSlideR23toRLongMax_RegionA(  rnew )
  endif
  rabAfterSlide = computeDistances(rnew, regA_Angles)

end function getDistancesAfterSlide_RegionA


function slideR13toRMedMax_RegionA(r12) result(rnew)
  implicit none
  double precision, intent(in)::  r12
  type(CentreToCentreTuple) :: rnew
  double precision xnew

  xnew= 0.5d0 *( cosThetaHat_A*r12 + Sqrt( r12**2*(-1.d0 + cosThetaHat_A**2) +  4.d0 * RMedMax**2) )
  rnew%r12 = r12
  rnew%r23 = Sqrt( r12**2/4.d0 + xnew*xnew + r12*xnew*cosThetaHat_A )
  rnew%r13 = RMedMax
  
end function slideR13toRMedMax_RegionA
  

function scaleThenSlideR23toRLongMax_RegionA(  rold ) result (rnew)
  implicit none
  type(CentreToCentreTuple), intent(in):: rold
  type(CentreToCentreTuple) :: rnew
  double precision :: xnew, debt, scaleSize
  
  rnew = rold
  
  debt = rnew%r23 - RLongMax
  scaleSize=  rold%r12/( rold%r12-debt/2.d0)
  rnew%r12 = rold%r12 / scaleSize
  
  !slide so that r23=maximum value
  xnew= 0.5d0 *(-cosThetaHat_A * rnew%r12  +  &
       Sqrt( rnew%r12**2 * (-1.d0 + cosThetaHat_A**2) + 4.d0*RLongMax**2))

  !! make the slide
  if( xnew >0.0) then
     rnew%r23 = RLongMax
     rnew%r13 = Sqrt(rnew%r12**2 / 4.d0 + xnew*xnew  -  rnew%r12 * xnew*cosThetaHat_A)
  endif

end function scaleThenSlideR23toRLongMax_RegionA
  

  

double precision function energyWithScaleAndSlide( rab, rabAfterSlide)
  implicit none
  double precision, intent(in), dimension(:) :: rab, rabAfterSlide
  double precision, dimension(nDim) :: xStar , rabAtRef
  
  double precision rRef, rGP,rTrue, Apref, Bpref
  double precision scale1,E_ref,  E_GP
 

  !!======== GP calculation using slide geometry========
  scale1 = rab(5)/rabAfterSlide(5)
  rGP = minval(rabAfterSlide(10:12))
  xStar = 1.d0/rabAfterSlide
  E_GP = PES_GP( xStar)/scale1**9
  

  !! ========Reference point (r13=r13ref) ========
  rabAtRef = getRabAtReferencePoint( rab )
  xStar = 1.d0/rabAtRef
  E_ref = Energy_GP_RegionA( xStar)
  rRef = minval(rabAtRef(10:12))


  !!======== Compute the two prefactors========
  Apref= (E_GP*rRef**beta  -  E_ref*rGP**beta)/ &
       ( rRef**beta * rGP**alpha  -  rRef**alpha * rGP**beta)
  
  Bpref= (-E_GP*rRef**alpha  +  E_ref*rGP**alpha)/ &
       ( rRef**beta * rGP**alpha  -  rRef**alpha * rGP**beta)

  
  !======== True geometry========
  rTrue=minval(rab(10:12))
  
  energyWithScaleAndSlide =  Apref*rTrue**alpha + Bpref*rTrue**beta
  
end function energyWithScaleAndSlide


function getRabAtReferencePoint( rab )  result( rabAtRef )
  implicit none
  double precision, intent(in), dimension(:) :: rab
  double precision, dimension(nDim):: rabAtRef
  type( CentreToCentreTuple ) :: RCentres
  double precision :: xref !! x [distance between 3 and centre of 1,2]


  RCentres%r12 = rab(5)
  RCentres%r13 = r13ref
  
  xref = 0.5d0 * (   cosThetaHat_A * rab(5) + sqrt( rab(5)**2 * (cosThetaHat_A**2 - 1.d0) +&
       4.d0 * r13ref**2 ) )
  RCentres%r23 = sqrt( rab(5)**2/4.d0 + xref**2 + rab(5)*xref*cosThetaHat_A )

  
  rabAtRef = computeDistances(RCentres,regA_Angles)
  
  call assertEqual_double( rabAtRef(11) , r13ref , 1d-6, "Rabref has the wrong r13 distance")
end function getRabAtReferencePoint
    

function Energy_GP_RegionA(xStar)
  implicit none
    double precision :: Energy_GP_RegionA
  double precision, dimension(:) :: xStar
  integer i,j,k
  double precision kKernTotal, kSqExpJthPerm,  kSqExpAllPerms
  
  kKernTotal=0
  do i=1,nTraining_A_Energy
     kSqExpAllPerms=0
     do j=1,nPerms
        kSqExpJthPerm=1        
        do k=1,nDim
           kSqExpJthPerm  =  kSqExpJthPerm * &
                (exp( -(xStar(k)-xTrainingPerm_A_Energy(k,j,i))**2 /2.d0/lScale_A_Energy(k)**2))
        end do !Dimensions (k)
        kSqExpAllPerms = kSqExpAllPerms + kSqExpJthPerm
     end do !Permuations (j)
     kKernTotal = kKernTotal + alpha_A_Energy(i) * kSqExpAllPerms
  end do !Training points (i)
  
  Energy_GP_RegionA=kKernTotal * expVar_A_Energy
  return
  
end function Energy_GP_RegionA


end module RegionA_Extrapolate_Module
