module RegionD_Extrapolate_Module
  use Assert_Module
  use PES_GP_Module, only: PES_GP
  use GP_variables_2CO2_Ar, only: nDim,nPerms, lCO
  use compDist_2CO2_Ar, only: computeDistances, compX, compThetaHat, compThetaTilde, &
       findSmallestDeltaXCO2Ar, findSmallestDeltaX2CO2,computedistances_withco2slide,&
       computeCtoO3, computeCtoO1
  use compAngles_2CO2_Ar, only:computeAngles
  use GP_variables_Asympt_2CO2_Ar !Data structure specific to this region so need all variables
  use RegionD_Options_2CO2_Ar !Data structure specific to this region so need all variables
  use CentreToCentreTuple_Module, only: CentreToCentreTuple
  use AnglesStructures, only: Angles
  implicit none

  !!====Public interface================================================
  public RegionD_Extrapolate
  !!====================================================================

  !!====Public variables/ functions - for tests only==================== 
  double precision, public :: cosThetaTilde_D,  x_D
  type( Angles ), public :: regD_Angles
  
  public getDistancesAfterSlide_RegionD, Energy_GP_RegionD
  !!====================================================================

  private

contains

  double precision function RegionD_Extrapolate( rab )
    implicit none
    double precision, intent(in), dimension(:) :: rab
    double precision, dimension(nDim) :: xStar

    if( minval(rab)<2.d0) then  ! return the repulsive part
       RegionD_Extrapolate = 0.0
       return
    endif

    if(isInsideGeometricConstrantRegionD( rab ) ) then
       xStar(:) = 1/rab(:)
       RegionD_Extrapolate = PES_GP( xStar)
       return
    endif

    RegionD_Extrapolate = RegionD_Extrapolate_WithSlide( rab )
    
  end function RegionD_Extrapolate

  
  logical function isInsideGeometricConstrantRegionD( rab )
    implicit none
    double precision, intent(in), dimension(:) :: rab
    
    if(minval(rab(10:12))< rShortMax .and.  minval(rab(1:9))< RMedMax  &
         .and. minval(rab(13:15))< RLongMax) then
       isInsideGeometricConstrantRegionD = .true.
    else
       isInsideGeometricConstrantRegionD = .false.
    end if
  end function isInsideGeometricConstrantRegionD


  double precision function RegionD_Extrapolate_WithSlide( rab )
    implicit none
    double precision, intent(in), dimension(nDim) :: rab
    double precision , dimension(nDim) :: rabAfterSlide
    
    rabAfterSlide = getDistancesAfterSlide_RegionD(rab)

    RegionD_Extrapolate_WithSlide = energyWithScaleAndSlide(rab, rabAfterSlide)

  end function RegionD_Extrapolate_WithSlide

  
  function getDistancesAfterSlide_RegionD(rab) result (rabAfterSlide )
    implicit none
    double precision, intent(in), dimension(:) :: rab
    double precision, dimension(nDim):: rabAfterSlide

    regD_Angles = computeAngles( rab)
    x_D= compX( rab(11), rab(5), rab(14)) !!! Order of the arguments depends on the region
    cosThetaTilde_D = compThetaTilde(rab(5), rab(11), x_D)
    
    rabAfterSlide = slideR12toRMedMax_RegionD(rab)

    if( minval(rabAfterSlide(13:15)) > RLongMax) then
       rabAfterSlide = scaleThenSlideR23toRLongMax_RegionD( rab, rabAfterSlide)
    endif

  end function getDistancesAfterSlide_RegionD

  
  function slideR12toRMedMax_RegionD(rab) result (rabAfterSlide )
    implicit none
    double precision, intent(in), dimension(:) :: rab
    double precision, dimension(nDim):: rabAfterSlide
    type(CentreToCentreTuple) :: rnew
    double precision O1pos(3),O3pos(3),C1toC2(3),slideDirection(3),xnew

    O1pos = computeCtoO1( regD_Angles )
    O3pos = computeCtoO3( regD_Angles )
    C1toC2 = (/0.d0 , 0.d0 , rab(5) /)
    slideDirection = (/ sqrt(abs(1.d0-cosThetaTilde_D**2)), 0.d0 , -cosThetaTilde_D/)

    xnew = x_D
    if(minval(rab(1:9)) > RMedMax) then
       xnew = x_D - findSmallestDeltaX2CO2(O1pos, O3pos,  C1toC2 ,slideDirection , RMedMax )
    endif
    
    rnew = CentreToCentreTuple( rab(5), rab(11), rab(14) )
        
    rabAfterSlide = computeDistances_withCO2slide(rnew , regD_Angles, cosThetaTilde_D, x_D-xnew)

  end function slideR12toRMedMax_RegionD

  
  function scaleThenSlideR23toRLongMax_RegionD(  rab, rabAfterSlide ) result (rOut)
    implicit none
    double precision, intent(in), dimension(nDim) :: rab
    double precision, intent(in), dimension(nDim) :: rabAfterSlide
    double precision, dimension(nDim) :: rOut
    type(CentreToCentreTuple) :: rnew
    double precision :: debt, scale1
    double precision :: deltaX, O3pos(3), ArToC2(3), slideDirection(3)

    !! Compute the deficit and scale to recover just less tahn 1/2 of this
    debt = minval(rabAfterSlide(13:15))-RLongMax
    scale1=  minval(rabAfterSlide(13:15))/(  minval(rabAfterSlide(13:15))-debt/1.8)
    rnew = CentreToCentreTuple( rab(5)/scale1, rab(11)/scale1, rab(14)/scale1)
    
    !slide the remaining distance so that r23=maximum value
    O3pos = computeCtoO3( regD_Angles)
    slideDirection = (/ sqrt(abs(1.d0-cosThetaTilde_D**2)), 0.d0 ,  -cosThetaTilde_D /)

    ArToC2 = (/ 0.d0, 0.d0,  rnew%r12 /) - &
         rnew%r13*(/ regD_Angles%sinTheta23, 0.d0, regD_Angles%cosTheta23 /)
    
    deltaX = findSmallestDeltaXCO2Ar( O3pos , ArToC2 , slideDirection , RLongMax )
    deltaX = max( 0d0, deltaX) !! Ensure the slide is postive
    rOut = computeDistances_withCO2slide(rnew , regD_Angles , cosThetaTilde_D , deltaX)

    call assertEqual_double( minval(rOut(13:15)), RLongMax, 1d-6, &
         "scaleThenSlideRegionD r23 wrong")
  end function scaleThenSlideR23toRLongMax_RegionD
    
  
  double precision function energyWithScaleAndSlide( rab, rabAfterSlide)
    implicit none
    double precision, intent(in), dimension(:) :: rab,rabAfterSlide
    
    double precision,dimension(nDim) :: xStar, rabAtRef
    double precision rRef, rGP, rTrue, Apref, Bpref
    double precision scale1, E_ref, E_GP

    
    !!======== GP calculation using slide geometry========
    scale1 = rab(11)/rabAfterSlide(11)   
    rGP = minval( rabAfterSlide(1:9) )
    xStar = 1.d0/rabAfterSlide
    E_GP = PES_GP( xStar)/scale1**9
    

    !======== Reference point (r12=r12ref)========
    rabAtRef = getRabAtReferencePoint( rab ) 
    rRef= minval( rabAtRef(1:9) )
    xStar = 1.d0/rabAtRef
    E_ref = Energy_GP_RegionD( xStar )


    !!======== Compute the two prefactors========
    Apref= ( E_GP*rRef**beta - E_ref*rGP**beta) /  &
         (rRef**beta * rGP**alpha - rRef**alpha * rGP**beta)
    
    Bpref= (-E_GP*rRef**alpha+ E_ref*rGP**alpha)/  &
         (rRef**beta * rGP**alpha - rRef**alpha * rGP**beta)
    

    !======== True geometry========
    rTrue = minval( rab(1:9) )

    energyWithScaleAndSlide = Apref * rTrue**alpha  +  Bpref * rTrue**beta

  end function energyWithScaleAndSlide

  
  function getRabAtReferencePoint( rab )  result( rabAtRef )
    implicit none
    double precision, intent(in), dimension(:) :: rab
    double precision, dimension(nDim):: rabAtRef

    type(CentreToCentreTuple) :: R_Ref
    double precision :: xref !! x [distance between 3 and centre of 1,2]
    double precision cosThetaHat !![angle between Ar3 and centre of 1,2]
    
    cosThetaHat = compThetaHat(rab(11), rab(5), rab(14), x_D) !!Argument order depends on region
    xref = 0.5d0 *(   cosThetaHat *  rab(11)    + sqrt( rab(11)**2 * (cosThetaHat**2 - 1.d0) &
         + 4.d0 * r12ref**2 ) )
    R_Ref =  CentreToCentreTuple(rab(5),rab(11),rab(14) )
    rabAtRef = computeDistances_withCO2slide(R_Ref , regD_Angles , cosThetaTilde_D , x_D-xref)

    call assertEqual_double( rabAtRef(5) ,r12ref, 1d-8,"Rabref has the wrong r12 distance")
  end function getRabAtReferencePoint
    

  
  function Energy_GP_RegionD(xStar)
    implicit none
    double precision :: Energy_GP_RegionD
    double precision, dimension(:) :: xStar
    integer i,j,k
    double precision kKernTotal, kSqExpJthPerm,  kSqExpAllPerms

    kKernTotal=0
    do i=1,nTraining_D_Energy
       kSqExpAllPerms=0
       do j=1,nPerms
          kSqExpJthPerm=1        
          do k=1,nDim
             kSqExpJthPerm  =  kSqExpJthPerm * &
                  ( exp( - (xStar(k)-xTrainingPerm_D_Energy(k,j,i))**2 /2.d0/lScale_D_Energy(k)**2) )
          end do !Dimensions (k)
          kSqExpAllPerms = kSqExpAllPerms + kSqExpJthPerm
       end do !Permuations (j)
       kKernTotal = kKernTotal + alpha_D_Energy(i) * kSqExpAllPerms
    end do !Training points (i)

    Energy_GP_RegionD=kKernTotal * expVar_D_Energy
    return


  end function Energy_GP_RegionD

end module RegionD_Extrapolate_Module
