module RegionE_Extrapolate_Module
  use Assert_Module
  use PES_GP_Module, only: PES_GP
  use GP_variables_2CO2_Ar, only: nDim,nPerms, lCO
  use compDist_2CO2_Ar, only: computeDistances, compX, compThetaHat, compThetaTilde, &
       findSmallestDeltaXCO2Ar, findSmallestDeltaX2CO2,computedistances_withco2slide,&
       computeCtoO3, computeCtoO1
  use compAngles_2CO2_Ar, only:computeAngles
  use GP_variables_Asympt_2CO2_Ar !Data structure specific to this region so need all variables
  use RegionE_Options_2CO2_Ar !Data structure specific to this region so need all variables
  use CentreToCentreTuple_Module, only: CentreToCentreTuple
  use AnglesStructures, only: Angles
  implicit none

  !!====Public interface================================================
  public RegionE_Extrapolate
  !!====================================================================

  !!====Public variables/ functions - for tests only==================== 
  double precision, public :: cosThetaTilde_E, cosThetaHat_E, x_E
  type( Angles ), public :: regE_Angles
  
  public getDistancesAfterSlide_RegionE, Energy_GP_RegionE
  !!====================================================================

  private
  
contains

  double precision function RegionE_Extrapolate( rab )
    implicit none
    double precision, dimension(nDim), intent(in) :: rab
    double precision , dimension(nDim) :: xStar
    
    if( minval(rab)<2.d0) then     ! return the repulsive part
       RegionE_Extrapolate = 0.0
       return
    endif

    if(isInsideGeometricConstrantRegionE( rab ) ) then
       xStar(:) = 1/rab(:)
       RegionE_Extrapolate = PES_GP( xStar)
       return
    endif

    RegionE_Extrapolate = RegionE_Extrapolate_WithSlide( rab )
    
  end function RegionE_Extrapolate

  
  logical function isInsideGeometricConstrantRegionE( rab )
    implicit none
    double precision, intent(in), dimension(nDim) :: rab
    
    if(minval(rab(10:12))< rShortMax .and.  minval(rab(13:15))< RMedMax  &
         .and. minval(rab(1:9))< RLongMax) then
       isInsideGeometricConstrantRegionE = .true.
    else
       isInsideGeometricConstrantRegionE = .false.
    end if
  end function isInsideGeometricConstrantRegionE
  

  double precision function RegionE_Extrapolate_WithSlide( rab )
    implicit none
    double precision, intent(in), dimension(nDim) :: rab
    double precision, dimension(nDim):: rabAfterSlide
    
    rabAfterSlide = getDistancesAfterSlide_RegionE( rab )

    RegionE_Extrapolate_WithSlide = energyWithScaleAndSlide(rab, rabAfterSlide)
    
  end function RegionE_Extrapolate_WithSlide


  function getDistancesAfterSlide_RegionE(rab) result (rabAfterSlide )
    implicit none
    double precision, intent(in), dimension(nDim) :: rab
    double precision, dimension(nDim):: rabAfterSlide
   
    regE_angles = computeAngles( rab)
    x_E= compX( rab(11), rab(5), rab(14)) !!! Order of the arguments depends on region
    cosThetaHat_E = compThetaHat(rab(11), rab(5), rab(14), x_E)!Argument order depends on region
    cosThetaTilde_E = compThetaTilde(rab(5), rab(11), x_E)

    rabAfterSlide = slideR12toRMedMax_RegionE(rab)
    
    if( minval(rabAfterSlide(1:9)) > RLongMax) then
       rabAfterSlide = scaleThenSlideR12toRLongMax_RegionE( rab, rabAfterSlide)
    endif

  end function getDistancesAfterSlide_RegionE

  
  function slideR12toRMedMax_RegionE(rab) result (rabAfterSlide )
    implicit none
    double precision, intent(in), dimension(nDim) :: rab
    double precision, dimension(nDim):: rabAfterSlide
    type(CentreToCentreTuple) :: rnew
    double precision OC2(3), ArAr(3), ArToC2(3), slideDirection(3), xnew

    OC2 = computeCtoO3( regE_angles )
    ArAr = (/ 0.d0 , 0.d0 , 0.d0 /)
    ArToC2= (/-rab(11) *regE_Angles%sinTheta23, 0.d0 , rab(5)-rab(11) * regE_Angles%cosTheta23/)
    slideDirection =  (/ sqrt(abs(1.d0-cosThetaTilde_E**2)), 0.d0 , -cosThetaTilde_E /)
    rnew = CentreToCentreTuple( rab(5), rab(11), rab(14) )
    
    xnew = x_E
    if(minval(rab(13:15)) > RMedMax) then
       xnew = x_E - findSmallestDeltaX2CO2(ArAr, OC2, ArToC2, slideDirection, RMedMax )
    endif
    
    rabAfterSlide = computeDistances_withCO2slide(rnew, regE_Angles, cosThetaTilde_E, x_E-xnew)

  end function slideR12toRMedMax_RegionE

  
  function scaleThenSlideR12toRLongMax_RegionE(  rab, rabAfterSlide ) result (rOut)
    implicit none
    double precision, intent(in), dimension(nDim) :: rab
    double precision, intent(in), dimension(nDim) :: rabAfterSlide
    double precision, dimension(nDim) :: rOut
    type(CentreToCentreTuple) :: rnew
    double precision :: debt, scale1
    double precision :: deltaX, OC1(3), OC2(3), C1toC2(3), slideDirection(3)
   
    !! Compute the deficit and scale to recover just less than 1/2 of this
    debt = minval(rabAfterSlide(1:9))-RLongMax
    scale1=  minval(rabAfterSlide(1:9))/(  minval(rabAfterSlide(1:9))-debt/1.8)
    rnew = CentreToCentreTuple(  rab(5)/scale1, rab(11)/scale1, rab(14)/scale1 ) 
    
    !slide the remaining distance so that r12=maximum value
    OC1 =  computeCtoO1( regE_Angles)
    OC2 =  computeCtoO3( regE_Angles)
    C1toC2 = (/ 0.d0 , 0.d0 , rnew%r12 /)
    slideDirection = (/ sqrt(abs(1.d0-cosThetaTilde_E**2)), 0.d0 , -cosThetaTilde_E /)
    
    deltaX = findSmallestDeltaX2CO2(OC1, OC2, C1toC2, slideDirection, RLongMax )
    deltaX = max( 0d0, deltaX) !! Ensure the slide is postive

    rOut = computeDistances_withCO2slide(rnew, regE_Angles, cosThetaTilde_E, deltaX)

    call assertEqual_double( minval(rOut(1:9)), RLongMax, 1d-6, &
         "scaleThenSlideRegionE r12 wrong")

  end function scaleThenSlideR12toRLongMax_RegionE
    
  
  double precision function energyWithScaleAndSlide( rab, rabAfterSlide)
    implicit none
    double precision, intent(in), dimension(nDim) :: rab,rabAfterSlide
    
    double precision, dimension(nDim) :: xStar, rabAtRef
    double precision rRef, rGP, rTrue, Apref, Bpref
    double precision scale1, E_ref, E_GP

    
    !!======== GP calculation using slide geometry========
    scale1 = rab(11)/rabAfterSlide(11)
    rGP= minval(rabAfterSlide(13:15))
    xStar = 1.d0/rabAfterSlide
    E_GP = PES_GP( xStar)/scale1**9


    !======== Reference point (r23=r23ref)========
    rabAtRef = getRabAtReferencePoint( rab ) 
    xStar = 1.d0/rabAtRef
    E_ref = Energy_GP_RegionE( xStar)
    rRef = minval(rabAtRef(13:15))


    !!======== Compute the two prefactors========
    Apref= (E_GP*rRef**beta  -  E_ref*rGP**beta)/  &
         ( rRef**beta * rGP**alpha  -  rRef**alpha * rGP**beta)
    
    Bpref= (-E_GP*rRef**alpha  +  E_ref*rGP**alpha)/  &
         ( rRef**beta * rGP**alpha  -  rRef**alpha * rGP**beta)

    
    !======== True geometry========
    rTrue = minval(rab(13:15))
    energyWithScaleAndSlide = Apref*rTrue**alpha + Bpref*rTrue**beta
    
  end function energyWithScaleAndSlide

  
  function getRabAtReferencePoint( rab )  result( rabAtRef )
    implicit none
    double precision, intent(in), dimension(nDim) :: rab
    double precision, dimension(nDim):: rabAtRef

    type(CentreToCentreTuple) :: rRef
    double precision :: xref !! x [distance between 3 and centre of 1,2]

    xref = 0.5d0* ( -cosThetaHat_E *  rab(11) + sqrt( rab(11)**2 * (cosThetaHat_E**2 - 1.d0) &
         + 4.d0 * r23ref**2 ) )
    rRef =  CentreToCentreTuple(rab(5),rab(11),rab(14) )
    rabAtRef = computeDistances_withCO2slide(rRef, regE_Angles, cosThetaTilde_E, x_E-xref)

    call assertEqual_double( rabAtRef(14), r23ref, 1d-8, &
         "Slide 2 has failed as r23 distance is wrong")
  end function getRabAtReferencePoint

  function Energy_GP_RegionE(xStar)
    implicit none
    double precision :: Energy_GP_RegionE
    double precision, dimension(:) :: xStar
    integer i,j,k
    double precision kKernTotal, kSqExpJthPerm,  kSqExpAllPerms

    kKernTotal=0
    do i=1,nTraining_E_Energy
       kSqExpAllPerms=0
       do j=1,nPerms
          kSqExpJthPerm=1        
          do k=1,nDim
             kSqExpJthPerm  =  kSqExpJthPerm * &
                  ( exp( - (xStar(k)-xTrainingPerm_E_Energy(k,j,i))**2 /2.d0/lScale_E_Energy(k)**2) )
          end do !Dimensions (k)
          kSqExpAllPerms = kSqExpAllPerms + kSqExpJthPerm
       end do !Permuations (j)
       kKernTotal = kKernTotal + alpha_E_Energy(i) * kSqExpAllPerms
    end do !Training points (i)

    Energy_GP_RegionE=kKernTotal * expVar_E_Energy
    return


  end function Energy_GP_RegionE


end module RegionE_Extrapolate_Module
